
import io
import os.path
import subprocess
import re
import modulefinder

from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GtkSource
from gi.repository import Pango
from gi.repository import GLib

import wayround_org.utils.path
import wayround_org.utils.timer
import wayround_org.utils.gtk

import wayround_org.pyeditor.buffer
import wayround_org.pyeditor.module_commons


MODE_NAME = 'javascript'

SUPPORTED_MIME = ['application/javascript']

SUPPORTED_FNM = ['*.js']


FUNCTION_REGEXP = re.compile(
    r'function\s+.*?\(.*?\).*?{'
    )


class Buffer(wayround_org.pyeditor.module_commons.Buffer):

    @staticmethod
    def get_mode_interface():
        return ModeInterface


class View(wayround_org.pyeditor.module_commons.View):

    @staticmethod
    def get_language_name():
        return MODE_NAME

    def apply_spec_view_settings(self):
        self.view.set_auto_indent(True)
        self.view.set_draw_spaces(GtkSource.DrawSpacesFlags.ALL)
        self.view.set_highlight_current_line(True)
        self.view.set_indent_on_tab(True)
        self.view.set_indent_width(4)
        self.view.set_insert_spaces_instead_of_tabs(True)
        self.view.set_right_margin_position(80)
        self.view.set_show_line_marks(True)
        self.view.set_show_line_numbers(True)
        self.view.set_show_right_margin(True)
        self.view.set_smart_home_end(True)
        self.view.set_tab_width(4)

        return


class SourceMenu:

    def __init__(self, mode_interface):

        self.mode_interface = mode_interface
        self.main_window = mode_interface.main_window

        main_window = self.main_window

        source_me = Gtk.Menu()

        # source_toggle_comment_mi = Gtk.MenuItem.new_with_label(
        #     "Toggle Comment"
        #     )
        # source_comment_mi = Gtk.MenuItem.new_with_label("Comment")
        # source_uncomment_mi = Gtk.MenuItem.new_with_label("Uncomment")

        source_indent_mi = Gtk.MenuItem.new_with_label("Indent")
        source_indent_mi.add_accelerator(
            'activate',
            main_window.accel_group,
            Gdk.KEY_I,
            Gdk.ModifierType.CONTROL_MASK,
            Gtk.AccelFlags.VISIBLE
            )
        source_indent_mi.connect(
            'activate',
            self.on_indent_mi,
            False
            )

        source_dedent_mi = Gtk.MenuItem.new_with_label("Dedent")
        source_dedent_mi.add_accelerator(
            'activate',
            main_window.accel_group,
            Gdk.KEY_I,
            Gdk.ModifierType.CONTROL_MASK
            | Gdk.ModifierType.SHIFT_MASK,
            Gtk.AccelFlags.VISIBLE
            )
        source_dedent_mi.connect(
            'activate',
            self.on_indent_mi,
            True
            )

        source_pep8_mi = Gtk.MenuItem.new_with_label("Use pep8.py")
        source_pep8_mi.set_no_show_all(True)
        source_autopep8_mi = Gtk.MenuItem.new_with_label("Use autopep8.py")
        source_autopep8_mi.set_no_show_all(True)
        source_autopep8_mi.add_accelerator(
            'activate',
            main_window.accel_group,
            Gdk.KEY_F,
            Gdk.ModifierType.CONTROL_MASK
            | Gdk.ModifierType.SHIFT_MASK,
            Gtk.AccelFlags.VISIBLE
            )
        source_autopep8_mi.connect(
            'activate',
            self.on_source_autopep8_mi
            )

        edit_delete_line_mi = Gtk.MenuItem.new_with_label("Delete Line")
        edit_delete_line_mi.add_accelerator(
            'activate',
            main_window.accel_group,
            Gdk.KEY_D,
            Gdk.ModifierType.CONTROL_MASK,
            Gtk.AccelFlags.VISIBLE
            )
        edit_delete_line_mi.connect(
            'activate',
            self.on_edit_delete_line_mi
            )

        navigate_refresh_outline_mi = \
            Gtk.MenuItem.new_with_label("Refresh Outline")

        navigate_refresh_outline_mi.add_accelerator(
            'activate',
            main_window.accel_group,
            Gdk.KEY_R,
            Gdk.ModifierType.CONTROL_MASK,
            Gtk.AccelFlags.VISIBLE
            )
        navigate_refresh_outline_mi.connect(
            'activate',
            self.on_navigate_refresh_outline_mi
            )
        edit_delete_trailing_whitespace_mi = Gtk.MenuItem.new_with_label(
            "Delete Trailing Whitespace"
            )
        edit_delete_trailing_whitespace_mi.add_accelerator(
            'activate',
            main_window.accel_group,
            Gdk.KEY_F,
            Gdk.ModifierType.CONTROL_MASK
            | Gdk.ModifierType.SHIFT_MASK
            | Gdk.ModifierType.MOD1_MASK,
            Gtk.AccelFlags.VISIBLE
            )
        edit_delete_trailing_whitespace_mi.connect(
            'activate',
            self.on_delete_trailing_whitespace_mi
            )

        # source_me.append(source_toggle_comment_mi)
        # source_me.append(source_comment_mi)
        # source_me.append(source_uncomment_mi)
        # source_me.append(Gtk.SeparatorMenuItem())

        source_me.append(edit_delete_line_mi)
        source_me.append(Gtk.SeparatorMenuItem())

        source_me.append(source_indent_mi)
        source_me.append(source_dedent_mi)
        source_me.append(Gtk.SeparatorMenuItem())

        source_me.append(source_pep8_mi)
        source_me.append(source_autopep8_mi)
        source_me.append(edit_delete_trailing_whitespace_mi)
        source_me.append(Gtk.SeparatorMenuItem())

        source_me.append(navigate_refresh_outline_mi)

        self._source_me = source_me

        return

    def get_widget(self):
        return self._source_me

    def destroy(self):
        self.get_widget().destroy()
        return

    def on_source_autopep8_mi(self, mi):

        try:
            import autopep8
        except:
            logging.exception("Can't use autopep8")
        else:

            buff = self.main_window.current_buffer

            if buff is not None:

                b = buff.get_buffer()

                t = b.get_text(
                    b.get_start_iter(),
                    b.get_end_iter(),
                    False
                    )

                buff.save_state()

                t = autopep8.fix_code(
                    t,
                    options=autopep8.parse_args(
                        ['--aggressive', '--ignore', 'E123', '']
                        )
                    )

                b.set_text(t)

                buff.restore_state()

        return

    def on_edit_delete_line_mi(self, mi):
        b = self.main_window.current_buffer.get_buffer()
        wayround_org.pyeditor.module_commons.delete_selected_lines(b)
        return

    def on_navigate_refresh_outline_mi(self, mi):
        mi = self.main_window.mode_interface
        if mi is not None and hasattr(mi, 'outline'):
            mi.outline.reload()
        return

    def _get_selected_lines(self):
        b = self.main_window.current_buffer.get_buffer()
        return wayround_org.pyeditor.module_commons.get_selected_lines(b)

    def on_indent_mi(self, mi, de=False):
        b = self.main_window.current_buffer.get_buffer()
        wayround_org.pyeditor.module_commons.indent_buffer(b, de, 4)
        return

    def on_delete_trailing_whitespace_mi(self, mi):
        buff = self.main_window.current_buffer
        b = buff.get_buffer()

        t = b.get_text(
            b.get_start_iter(),
            b.get_end_iter(),
            False
            )

        buff.save_state()

        t = wayround_org.pyeditor.module_commons.delete_trailing_whitespace(t)

        b.set_text(t)

        buff.restore_state()
        return


class Outline(wayround_org.pyeditor.module_commons.Outline):

    def search(self, buff):

        res = {}

        t = buff.get_text(
            buff.get_start_iter(),
            buff.get_end_iter(),
            False
            )

        excluded_ranges = []

        comments = wayround_org.pyeditor.module_commons.find_c_comments(t)

        excluded_ranges += comments

        last_start = 0
        while True:

            i = FUNCTION_REGEXP.search(t, last_start)

            if not i:
                break

            m_start = i.start()
            m_end = i.end()

            last_start = m_start + 1

            i_r = range(m_start, m_end)

            in_excluded_ranges = False
            for j in excluded_ranges:

                if i_r.start in j or i_r.stop in j:
                    in_excluded_ranges = True
                    break

                if j.start in i_r or j.stop in i_r:
                    in_excluded_ranges = True
                    break

            if not in_excluded_ranges:
                line = buff.get_iter_at_offset(m_start).get_line()
                s = buff.get_iter_at_line(line)
                e = buff.get_iter_at_offset(m_end)

                t2 = buff.get_text(s, e, False)
                # t2 = t[m_start: m_end]

                # res[line] = t2.strip()
                res[line] = t2
                excluded_ranges.append(i_r)

        return res


class ModeInterface:

    @staticmethod
    def get_menu_name():
        return "JavaScript"

    def __init__(self, main_window):
        self.main_window = main_window

        self.source_menu = SourceMenu(self)
        self.view = View(self)

        self.outline = Outline(self)

        self.lang_mgr = GtkSource.LanguageManager.get_default()

        #self.completion = self.view.get_view_widget().get_completion()
        #self.completion_provider = SourceCompletionProvider()
        #res = self.completion.add_provider(self.completion_provider)
        #print("add_provider: {}".format(res))

        # self.completion.create_context(None)
        return

    def destroy(self):
        self.source_menu.destroy()
        self.view.destroy()
        return

    def get_menu(self):
        return self.source_menu.get_widget()

    def get_view_widget(self):
        return self.view.get_view_widget()

    def get_view_widget_sw(self):
        return self.view.get_view_widget_sw()

    def get_widget(self):
        return self.view.get_widget()

    def set_buffer(self, buff):

        if not isinstance(buff, Buffer):
            raise Exception(
                "`buff' must be an instance of "
                "wayround_org.pyeditor.modes.python.Buffer"
                )

        buff.set_mode_interface(self)
        buff.set_language(self.lang_mgr.get_language('js'))
        self.view.set_buffer(buff)
        self.outline.reload()
        return

    def get_view(self):
        return self.view


def indent(txt, de=False):
    return wayround_org.pyeditor.module_commons.indent_text(txt, de, 4)





class Buffer:

    """
    Interface for buffers
    """

    def __init__(self, filename=None):
        return

    def open(self, filename):
        return

    def save(self, filename=None):
        return

    def get_modified(self):
        return

    def set_modified(self, value):
        return

    def get_buffer(self):
        return

    def destroy(self):
        return

    def get_title(self):
        return

    def get_mode_interface(self):
        return





from gi.repository import Gtk

SUPPORTED_MIME = []
SUPPORTED_FNM = []

class View:

    def __init__(self, mode_interface):

        self.mode_interface = mode_interface
        self.main_window = mode_interface.main_window

        self.view = Gtk.Label("Dummy")

        self._main = self.view

        return

    def get_view_widget_sw(self):
        return None

    def get_view_widget(self):
        return self._main

    def get_widget(self):
        return self._main

    def destroy(self):
        self.get_widget().destroy()


class SourceMenu:

    def __init__(self, mode_interface):

        self.mode_interface = mode_interface
        self.main_window = mode_interface.main_window

        source_me = Gtk.Menu()

        source_dummy_mi = Gtk.MenuItem.new_with_label(
            "Dummy"
            )
        source_dummy_mi.set_sensitive(False)

        source_me.append(source_dummy_mi)

        self._source_me = source_me

        return

    def get_widget(self):
        return self._source_me

    def destroy(self):
        self.get_widget().destroy()
        return


class ModeInterface:

    def __init__(self, main_window):
        self.main_window = main_window
        self.source_menu = SourceMenu(self)
        self.view = View(self)
        return

    def destroy(self):
        self.source_menu.destroy()
        self.view.destroy()
        return

    def get_menu(self):
        return self.source_menu.get_widget()

    def get_menu_name(self):
        return "Dummy"

    def get_widget(self):
        return self.get_view_widget()

    def get_view_widget(self):
        return self.view.get_widget()

    def get_view_widget_sw(self):
        return None

    def set_buffer(self, buff):
        return




import io
import os.path
import subprocess
import re
import modulefinder

from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GtkSource
from gi.repository import Pango
from gi.repository import GLib

import wayround_org.utils.path
import wayround_org.utils.timer
import wayround_org.utils.gtk

import wayround_org.pyeditor.buffer

C_COMMENT_RE = re.compile(
    r'/\*.*?\*/',
    flags=re.M | re.S,
    )

CPP_COMMENT_RE = re.compile(
    r'//.*$'
    )


class Buffer(
        GObject.GObject,
        wayround_org.pyeditor.buffer.Buffer
        ):

    __gsignals__ = {
        'changed': (GObject.SIGNAL_RUN_FIRST, None, tuple())
        }

    @staticmethod
    def get_mode_interface():
        raise Exception(
            "You must override this method in your class."
            " It must return ModeInterface class (not it's instance)"
            )
        return ModeInterface

    def __init__(self, main_window, filename=None):

        super().__init__()

        self.state = {}
        self.mode_interface = None

        self.main_window = main_window
        self.filename = filename
        self._b = None

        if filename is not None:
            self.open(filename)

        return

    def open(self, filename):

        t = ''

        if os.path.isfile(filename):

            with open(filename, 'r') as f:
                t = f.read()

        if self._b is None:
            self._b = GtkSource.Buffer()
        self._b.set_text(t)
        self._b.set_modified(False)
        self._b.connect(
            'modified-changed',
            self.on_buffer_modified_changed
            )

        self.filename = filename

        self.emit('changed')

        return

    def reopen(self):
        return self.open(self.filename)

    def save(self, filename=None):

        ret = 0

        if filename is None:
            filename = self.filename

        filename = wayround_org.utils.path.abspath(filename)

        d = os.path.dirname(filename)

        if not os.path.isdir(d):
            try:
                os.makedirs(d)
            except:
                pass

            if not os.path.isdir(d):
                ret = 1

        if ret == 0:

            t = self._b.get_text(
                self._b.get_start_iter(),
                self._b.get_end_iter(),
                False
                )

            with open(filename, 'w') as f:
                f.write(t)

            self._b.set_modified(False)

            if self.mode_interface:
                self.mode_interface.outline.reload()

        return ret

    def get_modified(self):
        return self._b.get_modified()

    def set_modified(self, value):
        return self._b.set_modified(value)

    def get_buffer(self):
        return self._b

    def get_filename(self):
        return self.filename

    def destroy(self):
        super().destroy()
        return

    def get_title(self):
        return os.path.basename(self.filename)

    def set_mode_interface(self, mode_interface):
        self.mode_interface = mode_interface
        return

    def set_language(self, language):
        self._b.set_language(language)
        return

    def on_buffer_modified_changed(self, widget):
        self.emit('changed')
        return

    def get_config(self):
        return self.state

    def set_config(self, data):
        self.state = data
        self.restore_state()
        return

    def save_state(self):

        if self._b:

            m = self._b.get_insert()
            i = self._b.get_iter_at_mark(m)
            cp = i.get_offset()
            self.state['cursor-position'] = cp

            if self.main_window.current_buffer == self:
                sw = self.mode_interface.get_view_widget_sw()
                if sw is not None:
                    vsb = sw.get_vscrollbar()
                    if vsb is not None:
                        value = vsb.get_value()
                        self.state['v-scroll-pos'] = value
                
                if hasattr(self.mode_interface.view, 'outline_sw'):
                    sw = self.mode_interface.view.outline_sw
                    if sw is not None:
                        vsb = sw.get_vscrollbar()
                        if vsb is not None:
                            value = vsb.get_value()
                            self.state['outline-v-scroll-pos'] = value
                    
        return

    def restore_state(self):

        if self._b:

            if 'cursor-position' in self.state:
                cp = self.state['cursor-position']

                i = self._b.get_iter_at_offset(cp)

                self._b.place_cursor(i)

            GLib.idle_add(self.restore_state_idle)

        return

    def restore_state_idle(self):
        if self.main_window.current_buffer == self:

            if 'v-scroll-pos' in self.state:

                sw = self.main_window.source_view_sw
                if sw is not None:
                    vsb = sw.get_vscrollbar()
                    if vsb is not None:
                        value = self.state['v-scroll-pos']
                        vsb.set_value(value)

            if 'outline-v-scroll-pos' in self.state:
                
                if hasattr(self.mode_interface.view, 'outline_sw'):
                    sw = self.mode_interface.view.outline_sw

                    if sw is not None:
                        vsb = sw.get_vscrollbar()
                        if vsb is not None:
                            value = self.state['outline-v-scroll-pos']
                            vsb.set_value(value)

        return


class View:

    @staticmethod
    def get_language_name():
        raise Exception(
            "You must override this staticmethod in your class."
            " It must return single word string (language name)"
            )
        return 'python'

    def __init__(self, mode_interface):

        b = Gtk.Box.new(Gtk.Orientation.VERTICAL, 5)

        self.mode_interface = mode_interface
        self.main_window = mode_interface.main_window

        paned_h2 = Gtk.Paned.new(Gtk.Orientation.HORIZONTAL)
        self.paned_h2 = paned_h2

        font_desc = Pango.FontDescription.from_string(
            self.main_window.get_fixed_text_editor_font_desc()
            )
        outline_treeview = Gtk.TreeView()
        outline_treeview.set_activate_on_single_click(True)
        outline_treeview.connect(
            'row-activated',
            self.on_outline_treeview_row_activated
            )
        outline_treeview.override_font(font_desc)
        outline_treeview.set_model(Gtk.ListStore(str, str))
        outline_treeview.set_headers_visible(False)
        self.outline = outline_treeview

        _c = Gtk.TreeViewColumn()
        _r = Gtk.CellRendererText()
        _c.pack_start(_r, False)
        _c.add_attribute(_r, 'text', 0)
        # _c.set_title('Line')
        outline_treeview.append_column(_c)

        _c = Gtk.TreeViewColumn()
        _r = Gtk.CellRendererText()
        _c.pack_start(_r, False)
        _c.add_attribute(_r, 'markup', 1)
        # _c.set_title('Text')
        outline_treeview.append_column(_c)

        outline_treeview_sw = Gtk.ScrolledWindow()
        self.outline_sw = outline_treeview_sw
        outline_treeview_sw.set_overlay_scrolling(False)
        outline_treeview_sw.add(outline_treeview)

        self.view = GtkSource.View()

        self.view.override_font(font_desc)

        self.apply_spec_view_settings()

        sw = Gtk.ScrolledWindow()
        self._sw = sw
        sw.add(self.view)
        sw.set_overlay_scrolling(False)

        self._status_label = Gtk.Label()
        self._status_label.set_alignment(0, 0.5)
        self._status_label.set_selectable(True)

        # sw_f = Gtk.Frame()
        # sw_f.add(sw)

        b.pack_start(sw, True, True, 0)
        b.pack_start(self._status_label, False, True, 0)

        self._main = paned_h2

        # outline_treeview_sw_f = Gtk.Frame()
        # outline_treeview_sw_f.add(outline_treeview_sw)

        paned_h2.add1(b)
        paned_h2.add2(outline_treeview_sw)
        paned_h2.child_set_property(outline_treeview_sw, 'resize', False)
        paned_h2.child_set_property(outline_treeview_sw, 'shrink', True)

        self._signal_pointer = None
        self._completion_sig_point = None

        if not self.main_window.cfg.cfg.has_section(self.get_language_name()):
            self.main_window.cfg.cfg.add_section(self.get_language_name())

        p3_pos = self.main_window.cfg.cfg.getint(
            self.get_language_name(),
            'paned_pos',
            fallback=500
            )

        paned_h2.set_position(p3_pos)

        return

    def apply_spec_view_settings(self):
        return

    def get_view_widget_sw(self):
        return self._sw

    def get_view_widget(self):
        return self.view

    def get_widget(self):
        return self._main

    def destroy(self):

        p = self.paned_h2.get_position()
        self.main_window.cfg.cfg.set(
            self.get_language_name(),
            'paned_pos',
            str(p)
            )

        self.main_window.buffer_clip.save_config()

        if self._main:
            self._main.destroy()
        if self.view:
            self.view.destroy()
        if self._sw:
            self._sw.destroy()
        return

    def set_buffer(self, buff):

        b = self.view.get_buffer()

        if b is not None:
            if self._signal_pointer:
                b.disconnect(self._signal_pointer)

        b = buff.get_buffer()
        self.view.set_buffer(b)

        self._signal_pointer = b.connect(
            'notify::cursor-position',
            self.on_cursor_position
            )

        return

    def _refresh_status(self):
        b = self.view.get_buffer()

        itera = b.get_iter_at_mark(b.get_insert())

        self._status_label.set_text(
            """\
line index: {} | column index: {} | offset (hex): {:x}
line: {} | column: {} | offset: {}""".format(
                itera.get_line(),
                itera.get_line_offset(),
                itera.get_offset(),
                itera.get_line() + 1,
                itera.get_line_offset() + 1,
                itera.get_offset()
                )
            )

    def on_cursor_position(self, gobject, pspec):
        self._refresh_status()
        return

    def on_outline_treeview_row_activated(self, widget, path, column):

        v = self.view

        m = widget.get_model()
        line = int(m[path][0])

        if v:
            b = v.get_buffer()
            i = b.get_iter_at_line(line - 1)
            b.place_cursor(i)
            v.scroll_to_iter(i, 0, True, 0.0, 0.2)
        return


class Outline:

    def __init__(self, mode_interface):
        self.mode_interface = mode_interface
        self.main_window = mode_interface.main_window
        self.source_view = mode_interface.get_view_widget()
        self.outline = self.mode_interface.get_view().outline

    def clear(self):
        m = self.outline.get_model()

        chi = m.get_iter_first()
        res = True

        while chi is not None and res is not False:
            res = m.remove(chi)

        return

    def reload(self):

        val = None
        o_sw = self.mode_interface.get_view().outline_sw
        vscrl = o_sw.get_vscrollbar()
        if vscrl:
            val = vscrl.get_value()

        self.clear()

        m = self.outline.get_model()

        b = self.source_view.get_buffer()

        res = self.search(b)

        for i in sorted(list(res.keys())):
            m.append([str(i + 1), res[i]])

        if val is not None:
            vscrl = o_sw.get_vscrollbar()
            if vscrl:
                GLib.idle_add(self._restore_vscroll, vscrl, val)

        return

    def search(self):
        raise Exception(
            "You must override this method in your class."
            " It must return dict where key in int and value is str"
            )
        return 'python'

    def _restore_vscroll(self, vscrl, val):
        vscrl.set_value(val)
        return


def get_selected_lines(buff):
    """
    Returns numbers of selected lines

    the 3rd result value - is bool, indicatind whatever some text is actually
    selected or 1st and 2nd results are simply indicating line in which cursor
    is currently positioned
    """

    ret = None, None, False

    b = buff

    if b:

        has_selection = b.get_has_selection()

        if not has_selection:
            ins = b.get_insert()
            ins_it = b.get_iter_at_mark(ins)
            ins_it_line = ins_it.get_line()

            ret = ins_it_line, ins_it_line, False

        else:

            first = b.get_iter_at_mark(b.get_insert()).get_offset()
            last = b.get_iter_at_mark(b.get_selection_bound()).get_offset()

            if first > last:
                _x = last
                last = first
                first = _x
                del _x

            first_l = b.get_iter_at_offset(first).get_line()
            last_l = first_l

            last_line_off = b.get_iter_at_offset(last).get_line_offset()

            if last_line_off == 0:
                last_l = b.get_iter_at_offset(last).get_line() - 1
            else:
                last_l = b.get_iter_at_offset(last).get_line()

            if last_l < first_l:
                last_l = first_l

            ret = first_l, last_l, True

    return ret


def indent_text(txt, de=False, indent_size=4):

    indent_space = ' ' * indent_size

    lines = txt.splitlines()

    if not de:
        for i in range(len(lines)):
            if lines[i] != '':
                lines[i] = '{}{}'.format(indent_space, lines[i])
    else:
        can_dedent = True
        for i in lines:
            if not i.startswith(indent_space) and not i == '':
                can_dedent = False
                break
        if can_dedent:
            for i in range(len(lines)):
                if lines[i] != '':
                    lines[i] = lines[i][indent_size:]

    return '\n'.join(lines)


def indent_buffer(buff, de=False, indent_size=4):

    b = buff

    res = get_selected_lines(buff)

    do_final_select = res[2]

    f_i = b.get_iter_at_line(res[0])
    l_i = b.get_iter_at_offset(
        b.get_iter_at_line(
            res[1] + 1
            ).get_offset() - 1,
        )

    t = b.get_text(f_i, l_i, False)

    t = indent_text(t, de, indent_size)

    b.delete(f_i, l_i)

    b.insert(f_i, t)

    f_i = b.get_iter_at_line(res[0])
    l_i = b.get_iter_at_offset(
        b.get_iter_at_line(
            res[1] + 1
            ).get_offset() - 1,
        )

    if do_final_select:
        b.select_range(f_i, l_i)
    else:
        pass

    return


def delete_selected_lines(buff):

    res = get_selected_lines(buff)

    buff.delete(
        buff.get_iter_at_line(res[0]),
        buff.get_iter_at_line(res[1] + 1)
        )

    return


def find_c_comments(text):
    """
    Searches for comments and returns list of ranges

    `text' must be bytes
    """

    comments_list = []

    for i in C_COMMENT_RE.finditer(text):
        r = range(i.start(), i.end())
        comments_list.append(r)

    for i in CPP_COMMENT_RE.finditer(text):
        r = range(i.start(), i.end())
        comments_list.append(r)

    ret = merge_c_overlapping_comments(comments_list)

    return ret


def merge_c_overlapping_comments(comments_list):
    ret = []

    for i in comments_list:
        found = False
        for j in ret:

            if i.start in j or i.stop in j:

                j_index = ret.index(j)

                if i.start in j and not i.stop in j:
                    ret[j_index] = range(j.start, i.stop)
                    j = ret[j_index]

                if i.end in j and not i.start in j:
                    ret[j_index] = range(i.start, j.stop)
                    j = ret[j_index]

                found = True
                break

        if not found:
            ret.append(i)

    return ret


def delete_trailing_whitespace(text):
    lines = text.split('\n')
    for i in range(len(lines)):
        lines[i] = lines[i].rstrip()
    return '\n'.join(lines)




import collections
import configparser
import os.path

import wayround_org.utils.path


class Config:

    def __init__(self, main_window, cfg_file=None):
        self.cfg = configparser.ConfigParser()

        if cfg_file is None:
            cfg_file = wayround_org.utils.path.join(
                os.path.expanduser('~')
                )
            cfg_file = wayround_org.utils.path.join(
                cfg_file, '.config', 'PyEditor', 'config.ini'
                )

        self._cfg_file = cfg_file

        return

    def load(self):

        if os.path.isfile(self._cfg_file):
            self.cfg.read(self._cfg_file)
        else:
            self.cfg.read_dict(
                collections.OrderedDict([
                    ('general',
                     collections.OrderedDict([
                         ('projects', '[]')
                         ])
                     ),
                    ])
                )

        return

    def save(self):

        ret = 0

        filename = wayround_org.utils.path.abspath(self._cfg_file)

        d = os.path.dirname(filename)

        if not os.path.isdir(d):
            try:
                os.makedirs(d)
            except:
                pass

            if not os.path.isdir(d):
                ret = 1

        with open(filename, 'w') as f:
            self.cfg.write(f)

        return ret





import os.path

from gi.repository import Gtk, Gdk


class RenameFileDialog:

    def __init__(self, main_window, old_filename, new_filename_base=''):

        self.main_window = main_window
        self.old_filename = old_filename

        if new_filename_base == '':
            new_filename_base = os.path.basename(old_filename)

        self.result = None

        window = Gtk.Window(Gtk.WindowType.TOPLEVEL)
        window.set_title("Rename File")
        window.set_modal(True)
        window.set_transient_for(main_window.get_widget())
        window.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
        window.set_type_hint(Gdk.WindowTypeHint.DIALOG)
        self._window = window

        b = Gtk.Box.new(Gtk.Orientation.VERTICAL, 5)
        b.set_margin_top(5)
        b.set_margin_start(5)
        b.set_margin_end(5)
        b.set_margin_bottom(5)

        g = Gtk.Grid()
        g.set_column_spacing(5)
        g.set_row_spacing(5)
        g.set_row_homogeneous(True)

        b.pack_start(g, True, True, 0)

        window.add(b)

        g.attach(Gtk.Label("Current Full Filename"), 0, 0, 1, 1)
        g.attach(Gtk.Label("New base or relative name"), 0, 1, 1, 1)

        name_label = Gtk.Label(old_filename)
        name_label.set_hexpand(True)
        self._name_label = name_label

        newname_entry = Gtk.Entry()
        newname_entry.set_text(new_filename_base)

        newname_entry.set_hexpand(True)
        self._newname_entry = newname_entry

        g.attach(name_label, 1, 0, 1, 1)
        g.attach(newname_entry, 1, 1, 1, 1)

        bb = Gtk.ButtonBox.new(Gtk.Orientation.HORIZONTAL)

        ok_button = Gtk.Button("Ok")
        ok_button.connect('clicked', self.on_ok_button_clicked)
        cancel_button = Gtk.Button("Cancel")
        cancel_button.connect('clicked', self.on_cancel_button_clicked)
        self._cancel_button = cancel_button

        bb.pack_start(ok_button, False, False, 0)
        bb.pack_start(cancel_button, False, False, 0)

        b.pack_start(bb, False, False, 0)

        window.connect('delete-event', self.on_delete)

        return

    def run(self):
        self.show()
        Gtk.main()
        return self.result

    def show(self):
        self.get_widget().show_all()
        return

    def destroy(self):
        self.get_widget().destroy()
        return

    def on_delete(self, widget, event):
        return self._cancel_button.emit('clicked')

    def get_widget(self):
        return self._window

    def on_ok_button_clicked(self, widget):

        newname = self._newname_entry.get_text()

        if newname != '':

            self.result = newname
            self.stop()

        else:
            d = Gtk.MessageDialog(
                self.get_widget(),
                Gtk.DialogFlags.MODAL,
                Gtk.MessageType.ERROR,
                Gtk.ButtonsType.OK,
                "New name field must be not empty"
                )
            d.run()
            d.destroy()

        return

    def on_cancel_button_clicked(self, widget):
        self.result = None
        self.stop()
        return

    def stop(self):
        Gtk.main_quit()
        return




import collections
import os.path
import json

from gi.repository import GObject


class ProjectClip(GObject.GObject):

    __gsignals__ = {
        'list-changed': (GObject.SIGNAL_RUN_FIRST, None, tuple())
    }

    def __init__(self, main_window):

        self.projects = collections.OrderedDict()
        self.main_window = main_window

        super().__init__()

        return

    def add(self, name, path):

        if not os.path.isdir(path):
            raise TypeError("`proj' must be existing dir")

        if name not in self.projects:
            self.projects[name] = path
            self.main_window.cfg.cfg.set(
                'general',
                'projects',
                json.dumps(list(self.to_items()))
                )
            self.main_window.cfg.save()
            self.emit('list-changed')

        return

    def rm(self, name):
        ret = 0
        if name in self.projects:
            del self.projects[name]
            self.main_window.cfg.cfg.set(
                'general',
                'projects',
                json.dumps(list(self.to_items()))
                )
            self.main_window.cfg.save()
            self.emit('list-changed')
        return ret

    def load_config(self):

        res = self.main_window.cfg.cfg.get(
            'general',
            'projects'
            )

        if res is not None:
            self.projects = collections.OrderedDict(json.loads(res))

        self.emit('list-changed')

        return

    def get(self, name):
        ret = None

        if name in self.projects:
            ret = self.projects[name]

        return ret

    def get_dict(self):
        return self.projects

    def get_list(self):
        return sorted(list(self.projects.keys()))

    def to_items(self):
        return self.projects.items()

    def from_items(self, items):
        self.projects = collections.OrderedDict(items)
        return




import os.path

from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GLib

import wayround_org.utils.gtk
import wayround_org.utils.path
import wayround_org.utils.file
import wayround_org.utils.error

import wayround_org.pyeditor.rename_file_dialog


class ProjectMenu:

    def __init__(self, main_window, files_tree):

        if not isinstance(
                files_tree,
                wayround_org.utils.gtk.DirectoryTreeView
                ):
            raise TypeError(
                "`files_tree' must be instance of "
                "wayround_org.utils.gtk.DirectoryTreeView"
                )

        self.main_window = main_window
        self.files_tree = files_tree

        menu = Gtk.Menu()

        open_file_manager_mi = Gtk.MenuItem("Open With External Program")
        open_file_manager_mi.connect('activate', self.on_open_file_manager_mi)

        create_directory_mi = Gtk.MenuItem("Create Dir..")
        create_directory_mi.connect('activate', self.on_create_directory_mi)

        rename_mi = Gtk.MenuItem("Rename..")
        rename_mi.connect('activate', self.on_rename_mi)

        delete_mi = Gtk.MenuItem("Delete..")
        delete_mi.connect('activate', self.on_delete_mi)

        menu.append(open_file_manager_mi)
        menu.append(Gtk.SeparatorMenuItem())
        menu.append(create_directory_mi)
        menu.append(Gtk.SeparatorMenuItem())
        menu.append(rename_mi)
        menu.append(Gtk.SeparatorMenuItem())
        menu.append(delete_mi)

        menu.show_all()

        self._main = menu

        return

    def get_widget(self):
        return self._main

    def on_open_file_manager_mi(self, mi):
        path = self.main_window.project_treeview.get_selected_path()

        path = 'file://' + path

        #path = GLib.uri_escape_string(path, None, True)

        print("path: {}".format(path))
        Gtk.show_uri(None, path, Gdk.CURRENT_TIME)
        return

    def on_create_directory_mi(self, mi):
        path = self.main_window.project_treeview.get_selected_path()

        if not os.path.isdir(path):
            path_spl = wayround_org.utils.path.split(path)
            p = wayround_org.utils.path.join(path_spl[:-1])
            if os.path.isdir(p):
                path = p

        d = Gtk.FileChooserDialog(
            "Create Directory",
            self.main_window.get_widget(),
            Gtk.FileChooserAction.CREATE_FOLDER,
            [
                'Ok', Gtk.ResponseType.OK,
                'Cancel', Gtk.ResponseType.CANCEL
                ]
            )
        d.set_current_folder(path)
        res = d.run()
        filename = None
        if res == Gtk.ResponseType.OK:
            filename = d.get_filename()

            os.makedirs(filename)

        d.destroy()
        return

    def on_rename_mi(self, mi):

        path = self.main_window.project_treeview.get_selected_path()

        d = wayround_org.pyeditor.rename_file_dialog.RenameFileDialog(
            self.main_window,
            path,
            ''
            )
        res = d.run()
        d.destroy()
        if res is not None:

            d = os.path.dirname(path)
            dn = wayround_org.utils.path.join(d, res)
            try:
                os.rename(path, dn)
            except:

                t = wayround_org.utils.error.return_exception_info(
                    sys.exc_info()
                    )

                d = Gtk.MessageDialog(
                    self.main_window.get_widget(),
                    Gtk.DialogFlags.MODAL,
                    Gtk.MessageType.ERROR,
                    Gtk.ButtonsType.OK,
                    "Can't rename file\n\n{}".format(t)
                    )
                d.run()
                d.destroy()

            pth_spl = wayround_org.utils.path.split(path)[:-1]
            if len(pth_spl) > 0:
                pth = wayround_org.utils.path.join(pth_spl)

                self.main_window.project_treeview.load_dir(
                    self.main_window.project_treeview.
                    get_selected_iter_parent(),
                    pth
                    )
        return

    def on_delete_mi(self, mi):

        path = self.main_window.project_treeview.get_selected_path()

        d = Gtk.MessageDialog(
            self.main_window.get_widget(),
            Gtk.DialogFlags.MODAL,
            Gtk.MessageType.QUESTION,
            Gtk.ButtonsType.YES_NO,
            "Confirm deletion of {}".format(path)
            )
        res = d.run()
        d.destroy()
        if res == Gtk.ResponseType.YES:
            wayround_org.utils.file.remove_if_exists(path)

            pth_spl = wayround_org.utils.path.split(path)[:-1]
            if len(pth_spl) > 0:
                pth = wayround_org.utils.path.join(pth_spl)

                self.main_window.project_treeview.load_dir(
                    self.main_window.project_treeview.
                    get_selected_iter_parent(),
                    pth
                    )
        return




import os.path

from gi.repository import Gtk, Gdk, Pango

import wayround_org.pyeditor.add_project_dialog
import wayround_org.pyeditor.buffer

import wayround_org.utils.path


class MainMenu:

    def __init__(self, main_window):

        self.main_window = main_window

        self.mode_interface = None

        mb = Gtk.MenuBar()

        tools_me = Gtk.Menu()

        tools_mi = Gtk.MenuItem.new_with_label("Tools")
        tools_mi.set_submenu(tools_me)

        change_font_mi = Gtk.MenuItem.new_with_label("Change Font..")
        tools_me.append(change_font_mi)
        change_font_mi.connect('activate', self.on_change_font_mi)

        file_mi = Gtk.MenuItem.new_with_label("File")
        project_mi = Gtk.MenuItem.new_with_label("Project")
        source_mi = Gtk.MenuItem.new_with_label("Source")
        self.source_mi = source_mi
        navigate_mi = Gtk.MenuItem.new_with_label("Navigate")

        mb.append(file_mi)
        mb.append(tools_mi)
        mb.append(project_mi)
        mb.append(source_mi)
        mb.append(navigate_mi)

        file_me = Gtk.Menu()
        file_mi.set_submenu(file_me)

        file_open_mi = Gtk.MenuItem.new_with_label("New/Open..")
        file_open_mi.add_accelerator(
            'activate',
            main_window.accel_group,
            Gdk.KEY_O,
            Gdk.ModifierType.CONTROL_MASK,
            Gtk.AccelFlags.VISIBLE
            )
        file_open_mi.connect('activate', self.on_file_open_mi)

        file_save_mi = Gtk.MenuItem.new_with_label("Save")
        file_save_mi.add_accelerator(
            'activate',
            main_window.accel_group,
            Gdk.KEY_S,
            Gdk.ModifierType.CONTROL_MASK,
            Gtk.AccelFlags.VISIBLE
            )
        file_save_mi.connect('activate', self.on_file_save_mi)

        file_save_as_mi = Gtk.MenuItem.new_with_label("Save as..")
        file_save_as_mi.add_accelerator(
            'activate',
            main_window.accel_group,
            Gdk.KEY_S,
            Gdk.ModifierType.CONTROL_MASK
            | Gdk.ModifierType.SHIFT_MASK
            | Gdk.ModifierType.MOD1_MASK,
            Gtk.AccelFlags.VISIBLE
            )

        file_save_all_mi = Gtk.MenuItem.new_with_label("Save All")
        file_save_all_mi.add_accelerator(
            'activate',
            main_window.accel_group,
            Gdk.KEY_S,
            Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.SHIFT_MASK,
            Gtk.AccelFlags.VISIBLE
            )
        file_save_all_mi.connect('activate', self.on_file_save_all_mi)

        file_close_mi = Gtk.MenuItem.new_with_label("Close")
        file_close_mi.add_accelerator(
            'activate',
            main_window.accel_group,
            Gdk.KEY_W,
            Gdk.ModifierType.CONTROL_MASK,
            Gtk.AccelFlags.VISIBLE
            )
        file_close_mi.connect('activate', self.on_file_close_mi)

        file_close_all_mi = Gtk.MenuItem.new_with_label("Close All")
        file_close_all_mi.connect('activate', self.on_file_close_all_mi)

        file_me.append(file_open_mi)
        file_me.append(file_save_mi)
        file_me.append(file_save_as_mi)
        file_me.append(Gtk.SeparatorMenuItem())
        file_me.append(file_save_all_mi)
        file_me.append(Gtk.SeparatorMenuItem())
        file_me.append(file_close_mi)
        file_me.append(Gtk.SeparatorMenuItem())
        file_me.append(file_close_all_mi)

        project_me = Gtk.Menu()
        project_mi.set_submenu(project_me)

        project_add_mi = Gtk.MenuItem.new_with_label("Add..")
        project_rm_mi = Gtk.MenuItem.new_with_label("Remove")

        project_add_mi.connect('activate', self.on_project_add_mi)
        project_rm_mi.connect('activate', self.on_project_rm_mi)

        project_me.append(project_add_mi)
        project_me.append(project_rm_mi)

        source_mi.set_submenu()

        navigate_me = Gtk.Menu()
        navigate_mi.set_submenu(navigate_me)

        navigate_next_buff_mi = Gtk.MenuItem.new_with_label("Next Buffer")
        navigate_next_buff_mi.add_accelerator(
            'activate',
            main_window.accel_group,
            Gdk.KEY_Page_Down,
            Gdk.ModifierType.CONTROL_MASK,
            Gtk.AccelFlags.VISIBLE
            )
        navigate_next_buff_mi.connect(
            'activate',
            self.on_navigate_next_buff_mi
            )

        navigate_prev_buff_mi = Gtk.MenuItem.new_with_label("Previous Buffer")
        navigate_prev_buff_mi.add_accelerator(
            'activate',
            main_window.accel_group,
            Gdk.KEY_Page_Up,
            Gdk.ModifierType.CONTROL_MASK,
            Gtk.AccelFlags.VISIBLE
            )
        navigate_prev_buff_mi.connect(
            'activate',
            self.on_navigate_prev_buff_mi
            )

        navigate_me.append(navigate_prev_buff_mi)
        navigate_me.append(navigate_next_buff_mi)

        self._main = mb

        return

    def get_widget(self):
        return self._main

    def destroy(self):
        self.get_widget().destroy()
        return

    def on_file_open_mi(self, mi):

        path = self.main_window.project_treeview.get_selected_path()

        if not os.path.isdir(path):
            path_spl = wayround_org.utils.path.split(path)
            p = wayround_org.utils.path.join(path_spl[:-1])
            if os.path.isdir(p):
                path = p

        d = Gtk.FileChooserDialog(
            "Select File to Open (non-existing file can be selected)",
            self.main_window._window,
            Gtk.FileChooserAction.SAVE,
            [
                'Ok', Gtk.ResponseType.OK,
                'Cancel', Gtk.ResponseType.CANCEL
            ]
            )
        d.set_create_folders(True)
        d.set_current_folder(path)
        res = d.run()
        filename = None
        if res == Gtk.ResponseType.OK:
            filename = d.get_filename()

            self.main_window.open_file(filename)

        d.destroy()

        return

    def on_file_close_mi(self, mi):
        self.main_window.close_current_buffer()
        return

    def on_file_close_all_mi(self, mi):
        for i in self.main_window.buffer_clip.buffers[:]:
            self.main_window.close_buffer(i)
        return

    def on_file_save_mi(self, mi):
        if self.main_window.current_buffer is not None:
            self.main_window.current_buffer.save()
        return

    def on_file_save_all_mi(self, mi):

        for i in self.main_window.buffer_clip.buffers:
            if i.get_modified():
                i.save()

        return

    def on_project_add_mi(self, mi):
        d = wayround_org.pyeditor.add_project_dialog.AddProjectDialog(
            self.main_window
            )
        res = d.run()
        d.destroy()

        if res is not None:
            self.main_window.projects.add(res['name'], res['directory'])

        return

    def on_project_rm_mi(self, mi):

        sel = self.main_window.projects_listview.get_selection()

        if sel:

            model, itera = sel.get_selected()

            if itera:

                path = model.get_path(itera)

                name = model[path][0]

                d = Gtk.MessageDialog(
                    self.main_window.get_widget(),
                    Gtk.DialogFlags.MODAL,
                    Gtk.MessageType.QUESTION,
                    Gtk.ButtonsType.YES_NO,
                    "Confirm project removal: {}\n\n"
                    "(none files removed. removing "
                    "only from PyEditor config records)".format(name)
                    )
                res = d.run()
                d.destroy()
                if res == Gtk.ResponseType.YES:
                    self.main_window.projects.rm(name)

        return

    def on_navigate_next_buff_mi(self, mi):

        self.main_window.select_current_buffer_in_list()

        sel = self.main_window.buffer_listview.get_selection()

        model, sel_rows = sel.get_selected_rows()

        if len(sel_rows) == 1:
            sel_row = sel_rows[0]

            sel_row_id = sel_row[0]

            if sel_row_id == model.iter_n_children() - 1:
                buf_id = int(model[Gtk.TreePath([0])][0])
            else:
                buf_id = int(model[Gtk.TreePath([sel_row_id + 1])][0])

            buf = None
            for i in self.main_window.buffer_clip.buffers:
                if id(i) == buf_id:
                    buf = i
                    break

            if buf is not None:
                self.main_window.set_buffer(buf)

        if self.main_window.source_view is not None:
            self.main_window.source_view.grab_focus()

        return

    def on_navigate_prev_buff_mi(self, mi):

        self.main_window.select_current_buffer_in_list()

        sel = self.main_window.buffer_listview.get_selection()

        model, sel_rows = sel.get_selected_rows()

        if len(sel_rows) == 1:
            sel_row = sel_rows[0]

            sel_row_id = sel_row[0]

            if sel_row_id == 0:
                buf_id = int(
                    model[Gtk.TreePath([model.iter_n_children() - 1])][0]
                    )
            else:
                buf_id = int(model[Gtk.TreePath([sel_row_id - 1])][0])

            buf = None
            for i in self.main_window.buffer_clip.buffers:
                if id(i) == buf_id:
                    buf = i
                    break

            if buf is not None:
                self.main_window.set_buffer(buf)

        if self.main_window.source_view is not None:
            self.main_window.source_view.grab_focus()

        return

    def on_change_font_mi(self, mi):
        desc = Pango.FontDescription.from_string(
            self.main_window.get_fixed_text_editor_font_desc()
            )
        d = Gtk.FontChooserDialog(
            "Select Fixed Font", 
            self.main_window.get_widget()
            )
        d.set_font_desc(desc)
        res = d.run()
        if res == Gtk.ResponseType.OK:
            self.main_window.set_fixed_text_editor_font_desc(
                d.get_font_desc().to_string()
                )
        d.destroy()
        return



#!/usr/bin/python3

import subprocess
import os.path

cwd = os.path.dirname(os.path.abspath(__file__))

if os.path.islink(__file__):
    cwd = os.path.dirname(os.path.abspath(os.readlink(__file__)))

p1 = subprocess.Popen(['python3', './main.py'], cwd=cwd)

try:
    input()
except:
    pass

try:
    p1.kill()
except:
    pass

exit(0)




import json
import collections

from gi.repository import GObject
from gi.repository import Gtk

import wayround_org.pyeditor.buffer
import wayround_org.utils.path
import wayround_org.utils.gtk


class ConfigFileListLoadingProcessWindow:

    def __init__(self):
        window = Gtk.Window()
        #window.set_default_size(300, 10)
        window.set_resizable(False)
        window.set_decorated(False)
        window.set_deletable(False)
        window.set_position(Gtk.WindowPosition.CENTER)
        window.set_title("PyEditor")

        prog_bar = Gtk.ProgressBar()
        # prog_bar.set_margin_top(10)
        # prog_bar.set_margin_bottom(10)
        # prog_bar.set_margin_start(10)
        # prog_bar.set_margin_end(10)

        b = Gtk.Box()
        b.set_orientation(Gtk.Orientation.VERTICAL)

        b.set_margin_top(5)
        b.set_margin_bottom(5)
        b.set_margin_start(5)
        b.set_margin_end(5)
        b.set_spacing(5)

        l = Gtk.Label("loading buffers.. please wait.")
        l.set_alignment(0, 0.5)

        b.pack_start(l, False, False, 0)
        b.pack_start(prog_bar, False, False, 0)

        window.add(b)

        self._window = window
        self._prog_bar = prog_bar
        return

    def show(self):
        self.get_widget().show_all()
        return

    def get_widget(self):
        return self._window

    def set_fraction(self, value):
        return self._prog_bar.set_fraction(value)

    def destroy(self):
        self.get_widget().destroy()


class BufferClip(GObject.GObject):

    __gsignals__ = {
        'list-changed-add': (
            GObject.SIGNAL_RUN_FIRST,
            None,
            (object,)
            ),
        'list-changed-rm': (GObject.SIGNAL_RUN_FIRST, None, (object,)),
        'list-changed-edit': (
            GObject.SIGNAL_RUN_FIRST,
            None,
            (object,)
            )
    }

    def __init__(self, main_window):

        self.buffers = []
        self.main_window = main_window

        super().__init__()

        return

    def add(self, buff):

        if not isinstance(buff, wayround_org.pyeditor.buffer.Buffer):
            raise Exception(
                "`buff' must be an instance of "
                "wayround_org.pyeditor.buffer.Buffer"
                )

        self.buffers.append(buff)

        self.buffers.sort(key=lambda x: x.get_filename())

        self.save_config()

        buff.connect('changed', self.on_buffer_changed_edit)

        self.emit('list-changed-add', (id(buff), buff,))

        return

    def remove(self, buff):

        if not isinstance(buff, wayround_org.pyeditor.buffer.Buffer):
            raise Exception(
                "`buff' must be an instance of "
                "wayround_org.pyeditor.buffer.Buffer"
                )

        ret = 0

        if buff in self.buffers:
            buff_id = id(buff)
            del self.buffers[self.buffers.index(buff)]
            buff.destroy()
            self.save_config()
            self.emit('list-changed-rm', (buff_id,))

        return ret

    def save_config(self):

        cfg = collections.OrderedDict()

        for i in self.buffers:

            setting_name = wayround_org.utils.path.realpath(i.get_filename())

            cfg[setting_name] = i.get_config()

        if 'buffer_settings' not in self.main_window.cfg.cfg.sections():
            self.main_window.cfg.cfg.add_section('buffer_settings')

        self.main_window.cfg.cfg.set(
            'buffer_settings',
            'buffer_state',
            json.dumps(list(cfg.items()))
            )

        self.main_window.cfg.save()

        return

    def load_config(self):

        cfg = self.main_window.cfg.cfg.get(
            'buffer_settings',
            'buffer_state',
            fallback=None
            )

        if cfg is not None:

            cfg = collections.OrderedDict(json.loads(cfg))

            lck = list(cfg.keys())
            lck_l = len(lck)

            pw = ConfigFileListLoadingProcessWindow()
            pw.set_fraction(0)
            pw.show()

            for ii in range(lck_l):
                i = lck[ii]
                res = self.main_window.open_file(i, False)

                if res != 1:
                    res.set_config(cfg[i])

                if ii == 0:
                    pw.set_fraction(0)
                else:
                    pw.set_fraction(1.0 / (lck_l / ii))
                wayround_org.utils.gtk.process_events()

            pw.set_fraction(1)
            wayround_org.utils.gtk.process_events()
            pw.destroy()

        return

    def on_buffer_changed_edit(self, widg):
        self.emit('list-changed-edit', (id(widg), widg,))
        return




import os.path
import logging
import importlib
import importlib.util
import modulefinder
import fnmatch
import collections

import mimetypes

from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GtkSource
from gi.repository import Pango

import wayround_org.utils.gtk
import wayround_org.utils.path

import wayround_org.pyeditor.buffer_clip
import wayround_org.pyeditor.config
import wayround_org.pyeditor.main_menu
import wayround_org.pyeditor.project_clip
import wayround_org.pyeditor.project_menu
import wayround_org.pyeditor.modes.dummy


class MainWindow:

    def __init__(self):

        self.cfg = wayround_org.pyeditor.config.Config(self)
        self.cfg.load()

        self.mode_interface = None
        self.current_buffer = None
        self.projects = wayround_org.pyeditor.project_clip.ProjectClip(self)
        self.projects.connect('list-changed', self.on_projects_list_changed)
        self.open_projects = []

        self.accel_group = Gtk.AccelGroup()

        window = Gtk.Window()
        window.set_title("PyEditor")
        window.add_accel_group(self.accel_group)
        window.set_hide_titlebar_when_maximized(True)
        window.connect('delete-event', self.on_delete)

        self.main_menu = wayround_org.pyeditor.main_menu.MainMenu(self)
        buffer_clip = wayround_org.pyeditor.buffer_clip.BufferClip(self)
        buffer_clip.connect(
            'list-changed-edit', self.on_buffer_clip_list_changed_edit
            )
        buffer_clip.connect(
            'list-changed-rm', self.on_buffer_clip_list_changed_rm
            )
        buffer_clip.connect(
            'list-changed-add', self.on_buffer_clip_list_changed_add
            )
        self.buffer_clip = buffer_clip

        b = Gtk.Box.new(Gtk.Orientation.VERTICAL, 0)

        menu_bar = self.main_menu.get_widget()

        self.source_widget = None
        self.source_view = None
        self.source_view_sw = None

        buffer_listview = Gtk.TreeView()
        buffer_listview_sw = Gtk.ScrolledWindow()
        buffer_listview_sw.set_overlay_scrolling(False)
        buffer_listview_sw.add(buffer_listview)
        buffer_listview.set_activate_on_single_click(True)
        # buffer_listview.set_headers_visible(False)
        self.buffer_listview = buffer_listview
        buffer_listview.set_model(
            Gtk.ListStore(
                str,
                # corresponding buffer instance id (integer converted to str)
                # as Python id(object) result is too large for int
                str,  # project
                str,  # filebasename
                str,  # modified?
                str,  # subdir in project
                str  # real path
                )
            )
        m = buffer_listview.get_model()
        m.set_sort_func(1, buffer_list_sorter1, None)
        m.set_sort_func(2, buffer_list_sorter2, None)
        del(m)

        buffer_listview.connect(
            'row-activated',
            self.on_buffer_listview_row_activated
            )

        _c = Gtk.TreeViewColumn()
        _r = Gtk.CellRendererText()
        _c.pack_start(_r, False)
        _c.add_attribute(_r, 'text', 1)
        _c.set_title('Project')
        _c.set_sort_column_id(1)
        buffer_listview.append_column(_c)

        _c = Gtk.TreeViewColumn()
        _r = Gtk.CellRendererText()
        _c.pack_start(_r, False)
        _c.add_attribute(_r, 'text', 2)
        _c.set_title('Name')
        _c.set_sort_column_id(1)
        buffer_listview.append_column(_c)

        _c = Gtk.TreeViewColumn()
        _r = Gtk.CellRendererText()
        _c.pack_start(_r, False)
        _c.add_attribute(_r, 'text', 3)
        _c.set_title('Changed?')
        _c.set_sort_column_id(2)
        buffer_listview.append_column(_c)

        _c = Gtk.TreeViewColumn()
        _r = Gtk.CellRendererText()
        _c.pack_start(_r, False)
        _c.add_attribute(_r, 'text', 4)
        _c.set_title('Display Path')
        _c.set_sort_column_id(1)
        buffer_listview.append_column(_c)

        _c = Gtk.TreeViewColumn()
        _c.set_visible(False)
        _r = Gtk.CellRendererText()
        _c.pack_start(_r, False)
        _c.add_attribute(_r, 'text', 5)
        _c.set_title('Path')
        buffer_listview.append_column(_c)

        projects_listview = Gtk.TreeView()
        self.projects_listview = projects_listview
        # projects_listview.set_headers_visible(False)
        projects_listview.set_model(Gtk.ListStore(str))
        projects_listview.connect(
            'row-activated',
            self.on_projects_listview_row_activated
            )

        _c = Gtk.TreeViewColumn()
        _r = Gtk.CellRendererText()
        _c.pack_start(_r, False)
        _c.add_attribute(_r, 'text', 0)
        _c.set_title('Name')
        projects_listview.append_column(_c)

        project_treeview = wayround_org.utils.gtk.DirectoryTreeView()
        self.project_treeview = project_treeview
        project_treeview.connect(
            'row-activated',
            self.on_project_treeview_row_activated
            )

        paned_v = Gtk.Paned.new(Gtk.Orientation.VERTICAL)
        # paned_v.set_property('handle-size', 5)
        self.paned_v = paned_v
        paned_h1 = Gtk.Paned.new(Gtk.Orientation.HORIZONTAL)
        self.paned_h1 = paned_h1

        projects_notebook = Gtk.Notebook()
        self.projects_notebook = projects_notebook

        projects_listview_sw = Gtk.ScrolledWindow()
        projects_listview_sw.set_overlay_scrolling(False)
        projects_listview_sw.add(projects_listview)

        projects_notebook.append_page(
            projects_listview_sw,
            Gtk.Label("Projects")
            )

        project_treeview_sw = Gtk.ScrolledWindow()
        project_treeview_sw.set_overlay_scrolling(False)
        project_treeview_sw.add(project_treeview)

        self.project_label = Gtk.Label("Dbl. click one of projects")

        projects_notebook.append_page(
            project_treeview_sw,
            self.project_label
            )

        projects_notebook.child_set_property(
            projects_listview_sw,
            'tab-expand',
            False
            )

        projects_notebook.child_set_property(
            project_treeview_sw,
            'tab-expand',
            True
            )

        self.project_menu = wayround_org.pyeditor.project_menu.ProjectMenu(
            self,
            self.project_treeview
            )

        self.project_treeview.connect(
            'button-press-event',
            self.on_project_treeview_button_press_event
            )

        # buffer_listview_sw_f = Gtk.Frame()
        # buffer_listview_sw_f.add(buffer_listview_sw)

        # projects_notebook_f = Gtk.Frame()
        # projects_notebook_f.add(projects_notebook)

        projects_notebook_sw = Gtk.ScrolledWindow()
        projects_notebook_sw.set_overlay_scrolling(False)
        projects_notebook_sw.add(projects_notebook)

        paned_v.add1(buffer_listview_sw)
        paned_v.add2(projects_notebook_sw)

        paned_h1.add1(paned_v)

        b.pack_start(menu_bar, False, False, 0)
        b.pack_start(paned_h1, True, True, 0)

        window.add(b)

        mxzd = self.cfg.cfg.get('general', 'maximized', fallback=True)

        w = self.cfg.cfg.getint('general', 'width', fallback=640)
        h = self.cfg.cfg.getint('general', 'height', fallback=480)

        p1_pos = self.cfg.cfg.getint('general', 'paned1_pos', fallback=-100)

        paned_v.set_position(p1_pos)

        p2_pos = self.cfg.cfg.getint('general', 'paned2_pos', fallback=100)

        paned_h1.set_position(p2_pos)

        # print('mxzd {}, w {}, h {}'.format(mxzd, w, h))

        if not mxzd:
            window.unmaximize()

        window.resize(w, h)

        if mxzd:
            window.maximize()

        self._window = window

        return

    def get_widget(self):
        return self._window

    def show(self):
        self.get_widget().show_all()
        return

    def destroy(self):
        self.main_menu.destroy()
        return

    def set_view_widget(self, source_widget, view_widget, view_widget_sw=None):
        self.source_widget = source_widget
        self.source_view = view_widget
        self.source_view_sw = view_widget_sw

        self.paned_h1.add2(source_widget)
        source_widget.show_all()
        return

    def open_file(
            self,
            filename,
            set_buff=True,
            force_mode=None
            ):
        ret = 0

        filename = wayround_org.utils.path.realpath(filename)
        # print("filename: {}".format(filename))

        mode = MODES['dummy']

        if not force_mode:

            file_mime = mimetypes.guess_type(filename)

            if file_mime in MODES_MIME_MAP:
                len_MODES_MIME_MAP_fm = len(MODES_MIME_MAP[file_mime])

                if len_MODES_MIME_MAP_fm == 0:
                    pass
                elif len_MODES_MIME_MAP_fm == 1:
                    mode = MODES_MIME_MAP[file_mime][
                        list(MODES_MIME_MAP[file_mime].keys())[0]
                        ]
                else:
                    # TODO: create mode selection dialog
                    pass

            # print("mode by mime: {}".format(mode))

            # if mode not found by mume type, try find it by filemask
            if mode == MODES['dummy']:
                acceptable_mode_mods = []
                for i in MODES_FNM_MAP.keys():
                    if fnmatch.fnmatch(os.path.basename(filename), i):
                        len_MODES_FNM_MAP_fm = len(MODES_FNM_MAP[i])

                        if len_MODES_FNM_MAP_fm == 0:
                            pass
                        elif len_MODES_FNM_MAP_fm == 1:
                            mode = MODES_FNM_MAP[i][
                                list(MODES_FNM_MAP[i].keys())[0]
                                ]
                            break
                        else:
                            # TODO: create mode selection dialog
                            pass

            # print("mode by fm: {}".format(mode))

        else:
            try:
                mode = MODES[force_mode]
            except:
                logging.exception("error setting mode `{}'".format(force_mode))
                ret = 2

        if mode == MODES['dummy']:
            d = Gtk.MessageDialog(
                self.get_widget(),
                Gtk.DialogFlags.MODAL,
                Gtk.MessageType.ERROR,
                Gtk.ButtonsType.OK,
                "Can't find suitable mode for file\n`{}'\n({})".format(
                    filename,
                    file_mime
                    )
                )
            d.run()
            d.destroy()
            ret = 1

        if ret == 0:

            buff = mode.Buffer(self)
            buff.open(filename)

            self.buffer_clip.add(buff)

            if set_buff:
                self.set_buffer(buff)

            ret = buff

        return ret

    def close_buffer(self, buff):

        ret = 0

        if buff in self.buffer_clip.buffers:

            if hasattr(buff, 'get_modified') and buff.get_modified() == True:
                d = Gtk.MessageDialog(
                    self.get_widget(),
                    Gtk.DialogFlags.MODAL,
                    Gtk.MessageType.QUESTION,
                    Gtk.ButtonsType.YES_NO,
                    "Buffer with Text for `{}' is not saved\n"
                    "Close this Buffer Anyway?".format(buff.get_title())
                    )
                res = d.run()
                d.destroy()
                if res == Gtk.ResponseType.YES:
                    ret = 0
                else:
                    ret = 1

        else:
            ret = 2

        if ret == 0:

            if buff == self.current_buffer:
                self.set_buffer(None)

            self.buffer_clip.remove(buff)
            buff.destroy()

        return

    def close_current_buffer(self):

        if (self.current_buffer is not None
                and self.current_buffer in self.buffer_clip.buffers):

            self.close_buffer(self.current_buffer)

        return

    def install_mode(self, name=None, cls=None):

        if name is not None:
            mod = load_mode(name)
            if isinstance(mod, int):
                logging.error("Can't load module")
            else:
                self.install_mode(cls=mod.ModeInterface)

        elif cls is not None:

            if not isinstance(self.mode_interface, cls):

                if self.mode_interface is not None:
                    self.mode_interface.destroy()

                mi = cls(self)

                self.mode_interface = mi

                menu = mi.get_menu()
                menu.show_all()

                self.main_menu.source_mi.set_submenu(menu)

                self.main_menu.source_mi.set_label(mi.get_menu_name())

                self.set_view_widget(
                    mi.get_widget(),
                    mi.get_view_widget(),
                    mi.get_view_widget_sw()
                    )

        return

    def set_buffer(self, buff):

        if buff is None or buff not in self.buffer_clip.buffers:
            self._window.set_title("PyEditor")

            if self.current_buffer is not None:
                self.current_buffer.save_state()

            self.install_mode('dummy')
            self.current_buffer = None
            self.select_current_buffer_in_list()

        else:

            if self.current_buffer is not None:
                self.current_buffer.save_state()

            self.install_mode(cls=buff.get_mode_interface())

            self.mode_interface.set_buffer(buff)

            self.current_buffer = buff

            self.select_current_buffer_in_list()

            self.current_buffer.restore_state()

            self._window.set_title(
                "{} - PyEditor".format(buff.get_title())
                )
        return

    def select_current_buffer_in_list(self):

        opened_index = -1

        buf_id = id(self.current_buffer)

        mod = self.buffer_listview.get_model()

        iter_ = None

        chi = mod.get_iter_first()

        if chi:
            while True:

                if int(mod[chi][0]) == buf_id:
                    iter_ = chi
                    break

                chi = mod.iter_next(chi)

                if not chi:
                    break

        if iter_ is not None:
            self.buffer_listview.get_selection().select_iter(iter_)

        return

    def signal_settings_changed(self):
        if hasattr(
                self.mode_interface, 'settings_changed'
                ):
            self.mode_interface.settings_changed()
        return

    def get_fixed_text_editor_font_desc(self):
        try:
            ret = self.cfg.cfg.get('general', 'fixed_text_editor_font_desc')
        except:
            ret = "Clean 9"
        return ret

    def set_fixed_text_editor_font_desc(self, desc):
        self.cfg.cfg.set('general', 'fixed_text_editor_font_desc', desc)
        self.cfg.save()
        self.signal_settings_changed()
        return

    def on_delete(self, widget, event):

        found_not_saved = False
        close_anyway = False
        ret = True

        for buff in self.buffer_clip.buffers:

            if hasattr(buff, 'get_modified') and buff.get_modified() == True:
                found_not_saved = True

        if found_not_saved:
            d = Gtk.MessageDialog(
                self.get_widget(),
                Gtk.DialogFlags.MODAL,
                Gtk.MessageType.QUESTION,
                Gtk.ButtonsType.YES_NO,
                "Not Saved Buffers found.\n"
                "Continue closing PyEditor?"
                )
            res = d.run()
            d.destroy()
            close_anyway = res == Gtk.ResponseType.YES

        if not found_not_saved or (found_not_saved and close_anyway):

            self.install_mode('dummy')

            self.cfg.cfg.set(
                'general',
                'maximized',
                str(self._window.is_maximized())
                )
            ws = self._window.get_size()
            self.cfg.cfg.set('general', 'width', str(ws[0]))
            self.cfg.cfg.set('general', 'height', str(ws[1]))

            self.cfg.cfg.set(
                'general',
                'paned1_pos',
                str(self.paned_v.get_position())
                )

            self.cfg.cfg.set(
                'general',
                'paned2_pos',
                str(self.paned_h1.get_position())
                )

            self.destroy()

            self.buffer_clip.save_config()
            Gtk.main_quit()
            ret = False
        return ret

    def on_buffer_clip_list_changed_add(self, widget, tup):

        id_, buff = tup
        m = self.buffer_listview.get_model()

        proj_dict = self.projects.get_dict()

        b_filename = wayround_org.utils.path.realpath(buff.get_filename())

        proj_name = ''

        for j, k in proj_dict.items():
            k_plus_slash = wayround_org.utils.path.realpath(k) + '/'
            if b_filename.startswith(k_plus_slash):
                proj_name = j
                break

        disp_file_path = b_filename

        if proj_name != '':
            disp_file_path = os.path.dirname(
                wayround_org.utils.path.relpath(
                    b_filename,
                    wayround_org.utils.path.realpath(proj_dict[proj_name])
                    )
                )

        m.append(
            [
                str(id_),
                proj_name,
                buff.get_title(),
                str(buff.get_modified()),
                disp_file_path,
                b_filename
                ]
            )
        return

    def on_buffer_clip_list_changed_edit(self, widget, tup):
        id_, buff = tup
        m = self.buffer_listview.get_model()

        b_filename = wayround_org.utils.path.realpath(buff.get_filename())

        proj_dict = self.projects.get_dict()

        proj_name = ''

        for j, k in proj_dict.items():
            k_plus_slash = wayround_org.utils.path.realpath(k) + '/'
            if b_filename.startswith(k_plus_slash):
                proj_name = j
                break

        chi = m.get_iter_first()
        if chi:
            while True:

                if int(m[chi][0]) == id_:
                    m.set_value(chi, 1, proj_name)
                    m.set_value(chi, 2, buff.get_title())
                    m.set_value(chi, 3, str(buff.get_modified()))

                chi = m.iter_next(chi)

                if chi in (False, None):
                    break

        return

    def on_buffer_clip_list_changed_rm(self, widget, tup):
        id_ = tup[0]
        m = self.buffer_listview.get_model()

        chi = m.get_iter_first()
        res = True

        if chi:
            while True:

                if int(m[chi][0]) == id_:
                    m.remove(chi)
                    break

                chi = m.iter_next(chi)

                if chi in (False, None):
                    break

        # while chi is not None and res is not False:
        #    res = m.remove(chi)
        return

    def on_projects_list_changed(self, widget):
        m = self.projects_listview.get_model()

        chi = m.get_iter_first()
        res = True

        while chi is not None and res is not False:
            res = m.remove(chi)

        lst = widget.get_list()

        for i in lst:
            m.append([i])

        return

    def on_buffer_listview_row_activated(self, widget, path, column):

        sel = widget.get_selection()
        model, iter_ = sel.get_selected()

        buf_id = int(model[iter_][0])

        buf = None
        for i in self.buffer_clip.buffers:
            if id(i) == buf_id:
                buf = i

        if buf is not None:
            self.set_buffer(buf)

            if self.source_view is not None:
                self.source_view.grab_focus()
        return

    def on_projects_listview_row_activated(self, widget, path, column):

        m = self.projects_listview.get_model()
        name = m[path][0]

        path = self.projects.get(name)
        self.project_treeview.set_root_directory(path)

        self.project_label.set_text(name)

        self.projects_notebook.set_current_page(1)

        return

    def on_project_treeview_row_activated(self, widget, path, column):
        pth = self.project_treeview.convert_indices_to_path(path.get_indices())
        fpth = wayround_org.utils.path.join(
            self.project_treeview.get_root_directory(),
            pth
            )

        if os.path.isfile(fpth):
            self.open_file(fpth)

        if self.source_view is not None:
            self.source_view.grab_focus()
        return

    def on_project_treeview_button_press_event(self, widget, event):
        if event.type == Gdk.EventType.BUTTON_PRESS and event.button == 3:
            self.project_menu.get_widget().popup(
                None, None, None, None, event.button, event.time
                )


def load_mode(name='dummy'):

    ret = 0

    if not isinstance(name, str):
        raise TypeError("`name' must be str")

    try:
        mod = importlib.import_module(
            'wayround_org.pyeditor.modes.{}'.format(name)
            )
    except:
        logging.exception("Can't load module `{}'".format(name))
        ret = 1
    else:
        ret = mod
    return ret


def find_modes():
    mf = modulefinder.ModuleFinder()
    ret = list(mf.find_all_submodules(wayround_org.pyeditor.modes))
    if 'text_plain' in ret:
        text_plain_pos = ret.index('text_plain')
        ret = ret[:text_plain_pos] + ret[text_plain_pos + 1:] + ['text_plain']
    # print("{}".format(ret))
    return ret


def create_module_map():

    ret = None, None, None

    mime_map = collections.OrderedDict()
    ext_map = collections.OrderedDict()
    modules = collections.OrderedDict()

    modes = find_modes()

    for i in modes:

        try:

            mod = importlib.import_module(
                'wayround_org.pyeditor.modes.{}'.format(i)
                )

        except:
            logging.exception("error loading mode module or package")
        else:

            if not hasattr(mod, 'SUPPORTED_MIME'):
                logging.error(
                    "mode module `{}' has not SUPPORTED_MIME attr".format(
                        mod
                        )
                    )
            else:

                for j in mod.SUPPORTED_MIME:
                    if not j in mime_map:
                        mime_map[j] = {}

                    if not mod in mime_map[j]:
                        mime_map[j][i] = mod

                modules[i] = mod

            if not hasattr(mod, 'SUPPORTED_FNM'):
                logging.error(
                    "mode module `{}' has not SUPPORTED_FNM attr".format(
                        mod
                        )
                    )
            else:

                for j in mod.SUPPORTED_FNM:
                    if not j in ext_map:
                        ext_map[j] = {}

                    if not mod in ext_map[j]:
                        ext_map[j][i] = mod

                modules[i] = mod

    ret = modules, mime_map, ext_map

    return ret


def buffer_list_sorter1(model, row1, row2, user_data):
    val11 = model[row1][1]
    val12 = model[row2][1]
    # print('val11: {}'.format(val11))
    # print('val12: {}'.format(val12))
    if val11 < val12:
        ret = -1
    elif val11 > val12:
        ret = 1
    else:
        val41 = model[row1][4]
        val42 = model[row2][4]
        # print('val21: {}'.format(val21))
        # print('val22: {}'.format(val22))
        if val41 < val42:
            ret = -1
        elif val41 > val42:
            ret = 1
        else:
            val21 = model[row1][2]
            val22 = model[row2][2]
            # print('val21: {}'.format(val21))
            # print('val22: {}'.format(val22))
            if val21 < val22:
                ret = -1
            elif val21 > val22:
                ret = 1
            else:
                ret = 0
    return ret


def buffer_list_sorter2(model, row1, row2, user_data):
    val31 = model[row1][3] == 'True'
    val32 = model[row2][3] == 'True'
    if val31 and not val32:
        ret = 1
    elif not val31 and val32:
        ret = -1
    else:
        ret = 0
    return ret

MODES, MODES_MIME_MAP, MODES_FNM_MAP = create_module_map()

# print("{}\n\n{}\n\n{}\n\n".format(MODES, MODES_MIME_MAP, MODES_FNM_MAP))



#!/usr/bin/python3

from setuptools import setup


setup(
    name='wayround_org_pyeditor',
    version='0.3.3',
    description='Simple extansible editor with projects and outline',
    author='Alexey V Gorshkov',
    author_email='animus@wayround.org',
    url='https://github.com/AnimusPEXUS/wayround_org_pyeditor',
    packages=[
        'wayround_org.pyeditor',
        'wayround_org.pyeditor.modes'
        ],
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: POSIX'
        ],
    entry_points={
        'console_scripts': 'pyeditor = wayround_org.pyeditor.main'
        },
    install_requires=['wayround_org_utils']
    )



#!/usr/bin/python3

import gi

gi.require_version('Gtk', '3.0')
gi.require_version('GtkSource', '3.0')

from gi.repository import Gtk

import wayround_org.pyeditor.main_window

w = wayround_org.pyeditor.main_window.MainWindow()

w.show()
w.install_mode('dummy')
w.projects.load_config()
w.buffer_clip.load_config()

Gtk.main()
